from strictyaml import Validator
from strictyaml.validators import MapValidator
from strictyaml.exceptions import YAMLSerializationError, InvalidValidatorError
from .control import Control
from .blocks import Block, Case, Overlay
from collections.abc import Callable
from .builder import ValidatorBuilder
from strictyaml.yamllocation import YAMLChunk


class DMap(MapValidator):
    def __init__(
        self,
        control: Control,
        blocks: list[Block],
        constraints: list[Callable[..., bool]] | None = None,
    ):
        assert isinstance(control, Control), "control must be of type Control"
        assert isinstance(blocks, list), "blocks must be a list of Block"
        for block in blocks:
            assert isinstance(block, Block), "all blocks must be of type Block"
            
        if constraints is not None:
            assert isinstance(constraints, list), "constraints must be a list of Callable"
            for constraint in constraints:
                assert callable(constraint), "every constraint must be callable"

        self.control = control
        self.blocks = blocks
        self.constraints = constraints

    @staticmethod
    def compile_when(when):
        if callable(when):
            return when
        return lambda raw, ctrl: bool(when)

    @staticmethod
    def compile_constraint(when):
        if callable(when):
            return when
        return lambda raw, ctrl, val: bool(when)

    @staticmethod
    def normalize_raw(raw):
        if isinstance(raw, dict):
            return {key: DMap.normalize_raw(value) for key, value in raw.items()}
        if isinstance(raw, list):
            return [DMap.normalize_raw(value) for value in raw]
        if isinstance(raw, str):
            lowered = raw.lower()
            if lowered == "true":
                return True
            if lowered == "false":
                return False
        return raw

    def validate(self, chunk):
        chunk.expect_mapping()
        self.control.validate(chunk)
        ctrl = self.control.validated.data
        raw = DMap.normalize_raw(chunk.whole_document)
        control_validator = self.control._validator
        true_case_block = None
        true_overlay_blocks = []
        # TODO: what if the user doesn't really want a control validator and only selects based on raw
        for block in self.blocks:
            if not DMap.compile_when(block.when)(raw, ctrl):
                continue
            if isinstance(block, Case):
                if true_case_block is None:
                    true_case_block = block
                else:
                    chunk.expecting_but_found("when evaluating DMap blocks", "multiple cases were true")
            elif isinstance(block, Overlay):
                true_overlay_blocks.append(block)
            else:
                chunk.expecting_but_found(
                    "when evaluating DMap blocks",
                    "unknown block type; expected Case or Overlay",
                )

        if true_case_block is None:
            chunk.expecting_but_found("when evaluating DMap blocks", "none of the cases were true")

        final_validator = ValidatorBuilder(
            control_validator,
            true_case_block._validator,
            [overlay._validator for overlay in true_overlay_blocks],
            self.control.source,
        ).validator

        self.validated = final_validator(chunk)
        val = self.validated.data

        if self.constraints:
            for constraint in self.constraints:
                if not DMap.compile_constraint(constraint)(raw, ctrl, val):
                    chunk.expecting_but_found("when evaluating DMap constraints", "constraints not fulfilled")
        if true_case_block.constraints:
            for constraint in true_case_block.constraints:
                if not DMap.compile_constraint(constraint)(raw, ctrl, val):
                    chunk.expecting_but_found("when evaluating DMap case constraints", "constraints not fulfilled")
        for overlay in true_overlay_blocks:
            if overlay.constraints:
                for constraint in overlay.constraints:
                    if not DMap.compile_constraint(constraint)(raw, ctrl, val):
                        chunk.expecting_but_found(
                            "when evaluating DMap overlay constraints",
                            "constraints not fulfilled",
                        )

    def to_yaml(self, data):
        self._should_be_mapping(data)
        self.control.validate(YAMLChunk(data))
        ctrl = self.control.validated.data
        raw = DMap.normalize_raw(data)

        true_case_block = None
        true_overlay_blocks = []
        for block in self.blocks:
            if not DMap.compile_when(block.when)(raw, ctrl):
                continue
            if isinstance(block, Case):
                if true_case_block is None:
                    true_case_block = block
                else:
                    raise YAMLSerializationError("Multiple DMap cases evaluated to true")
            elif isinstance(block, Overlay):
                true_overlay_blocks.append(block)
            else:
                raise YAMLSerializationError("Unknown DMap block type; expected Case or Overlay")

        if true_case_block is None:
            raise YAMLSerializationError("None of the DMap cases successfully serialized the data")

        final_validator = ValidatorBuilder(
            self.control._validator,
            true_case_block._validator,
            [overlay._validator for overlay in true_overlay_blocks],
            self.control.source,
        ).validator
        return final_validator.to_yaml(data)

    def __repr__(self):
        return "DMap({0}, {1}{2})".format(
            repr(self.control),
            repr(self.blocks),
            ", constraints={0}".format(repr(self.constraints)) if self.constraints else "",
        )



from strictyaml import Validator
from strictyaml.validators import MapValidator
from strictyaml.yamllocation import YAMLChunk
from functools import reduce
from strictyaml.ruamel.comments import CommentedMap
from strictyaml.exceptions import YAMLSerializationError


class Control:
    def __init__(self, validator: Validator, source: tuple[str] | str | None = None):
        self._validator = validator
        self.source = source
        self.validated = None

        assert isinstance(
            self._validator, Validator
        ), "validator must be of type Validator"

    def __repr__(self):
        return "Control({0}{1})".format(
            repr(self._validator),
            ", source={0}".format(repr(self.source)) if self.source else "",
        )

    def projection(self, chunk, validator):
        from .utils import unpack
        validator = unpack(validator)
        projected_chunk = {}

        if hasattr(validator, '_validator_dict'):
            keys = validator._validator_dict.items()
        elif hasattr(validator, '_validator') and isinstance(validator._validator, dict):
            # for when _validator_dict isn't explicitly built but _validator is a dict
            keys = [
                (k.key if hasattr(k, 'key') else k, v)
                for k, v in validator._validator.items()
            ]
        else:
            from strictyaml.exceptions import InvalidValidatorError
            raise InvalidValidatorError("Control validator must be a Map with specific keys")

        for key, val in keys:
            val = unpack(val)
            if key in chunk:
                if isinstance(val, MapValidator):
                    projected_chunk[key] = self.projection(
                        chunk[key], val
                    )
                else:
                    projected_chunk[key] = chunk[key]
                    
        return CommentedMap(projected_chunk)

    def validate(self, chunk):
        from .utils import unpack

        if self.source and self.source != "":
            if isinstance(self.source, str):
                chunk_pointer = chunk.contents[self.source]
            elif isinstance(self.source, tuple):
                chunk_pointer = reduce(
                    lambda d, key: d[key], self.source, chunk.contents
                )
        else:
            chunk_pointer = chunk.contents
        unpacked_validator = unpack(self._validator)
        is_mapping_validator = (
            hasattr(unpacked_validator, "_validator_dict")
            or (
                hasattr(unpacked_validator, "_validator")
                and isinstance(unpacked_validator._validator, dict)
            )
        )
        if is_mapping_validator:
            source_chunk = YAMLChunk(self.projection(chunk_pointer, self._validator))
        else:
            source_chunk = YAMLChunk(chunk_pointer)
        self.validated = self._validator(source_chunk)



from .forwardref import ForwardRef
from strictyaml import Map, MapCombined, MapPattern
from strictyaml.validators import Validator
from strictyaml.exceptions import YAMLSerializationError

def unpack(validator):
    while isinstance(validator, ForwardRef):
        if validator._validator is None:
            raise YAMLSerializationError("ForwardRef was used before it was set")
        validator = validator._validator
    return validator

def ensure_validator_dict(validator):
    validator = unpack(validator)
    if isinstance(validator, MapPattern):
        return MapCombined({}, validator._key_validator, validator._value_validator)
    return validator



from strictyaml import *

from .forwardref import ForwardRef
from .dmap import DMap
from .control import Control
from .blocks import Block, Case, Overlay
from .builder import ValidatorBuilder
from .utils import ensure_validator_dict, unpack



from strictyaml import Validator
from strictyaml.exceptions import YAMLSerializationError


class ForwardRefValidator(Validator):
    def _should_be_validator(self, validator):
        if not isinstance(validator, Validator):
            raise YAMLSerializationError(
                "Expected a Validator, found '{}'".format(validator)
            )


class ForwardRef(ForwardRefValidator):
    def __init__(self):
        self._validator = None
        self.has_expanded = False

    def set(self, validator: Validator):
        self._should_be_validator(validator)
        self._validator = validator

    def __call__(self, chunk):
        if self._validator is None:
            raise YAMLSerializationError("ForwardRef was used before it was set")
        return self._validator(chunk)

    def __repr__(self):
        if self._validator is None:
            return "{0}()".format(self.__class__.__name__)
        if not self.has_expanded:
            self.has_expanded = True
            return "{0}".format(repr(self._validator))
        else:
            return "{0}()".format(self.__class__.__name__)

    def to_yaml(self, data):
        return self._validator.to_yaml(data)
